# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

PVA object viewer utilities

@author: Guobao Shen <gshen@anl.gov>
"""

import numpy as np
import pyqtgraph
import pvaccess as pva
from ..model import ConnectionState
from .scope_controller_base import ScopeControllerBase
from ..view.scope_display import PlotChannel as ScopePlotChannel
from pyqtgraph.Qt import QtCore
import math
import statistics

class ScopeController(ScopeControllerBase):

    def __init__(self, widget, model, parameters, **kwargs):
        
        """

        :param model:
        :param parameters:
        :param channels:
        """
        self.color_pattern = kwargs.get("color", 
                                        ['#FFFF00', '#FF00FF', '#55FF55', '#00FFFF', '#5555FF',
                                         '#5500FF', '#FF5555', '#0000FF', '#FFAA00', '#000000'])

        nchannels = kwargs.get("channels", 4)
        self.channels = []
        for i in range(nchannels):
            self.channels.append(ScopePlotChannel('None', self.color_pattern[i]))

        warning = kwargs["WARNING"]

        super().__init__(widget, model, parameters, warning, channels=self.channels, **kwargs)
        
        self.default_arrayid = "None"
        self.default_xaxes = "None"
        self.current_arrayid = "None"
        self.current_xaxes = "None"

        #auto-set buffer size to waveform length.  This will be
        #disabled if buffer is set manually
        self.auto_buffer_size = True
        self.model.status_callback = self.connection_changed
        self.connection_timer = pyqtgraph.QtCore.QTimer()
        self.connection_timer.timeout.connect(self.__check_connection)
        class FlagSignal(QtCore.QObject):
            sig = QtCore.pyqtSignal(bool)
            def __init__(self):
                QtCore.QObject.__init__(self)

        self.connection_timer_signal = FlagSignal()
        self.connection_timer_signal.sig.connect(self.__failed_connection_callback)
        self.buffer_unit = 'Samples'
        self.object_size = None
        self.object_size_tally = []

    def default_config(self, **kwargs):
        """
        Update configuration based on commmand-line arguments
        Called in the top-level function and passes in the command-line arguments
        for scope as key-word arguments.
        Any options not set should be set in kwargs as None
        """


        start = self.parameters.child("Acquisition").child("Start").value()
        if kwargs['connect_on_start']:
            start = True

        bunit = self.parameters.child("Acquisition").child("Buffer Unit").value().strip()
        if bunit != "":
            self.buffer_unit = bunit
        
        super().default_config(**kwargs, buffer_unit=self.buffer_unit)
        self.auto_buffer_size = not self.parameters.child("Acquisition").child("Buffer (%s)" % self.buffer_unit).value()
        
        if kwargs['arrayid']:
            self.set_arrayid(kwargs["arrayid"])
        if kwargs['xaxes']:
            self.set_xaxes(kwargs['xaxes'])
            
        self._win.graphicsWidget.set_range(**kwargs)

        #Updates channel fields based on self.parameters
        self.update_fdr()

        if kwargs['fields']:
            fields = kwargs['fields'].split(',')
            for i, f in enumerate(fields):
                chan_name = "Channel %s" % (i + 1)
                child = self.parameters.child(chan_name)
                c = child.child("Field")
                c.setValue(f)
                self.set_channel_data(chan_name, 'Field', c.value())

        #Update other channel information
        for idx in range(len(self.channels)):
            chan_name = "Channel %s" % (idx + 1)
            child = self.parameters.child(chan_name)
            c = child.child("DC offset")
            if c.value() != 0:
                self.set_channel_data(chan_name, 'DC offset', c.value())
                
        child = self.parameters.child("Config").child("MO Disp Location")
        self._win.graphicsWidget.set_mouseover_display_location(child.value())

        # Apply Mouse Over setting from config
        mouse_over = self.parameters.child("Display").child("Mouse Over").value()
        self._win.graphicsWidget.set_enable_mouseover(mouse_over)

        # Apply Extra Display Fields from config
        extra_fields = self.parameters.child("Config").child("Extra Display Fields").value()
        if extra_fields:
            self._win.graphicsWidget.set_mouseover_fields(extra_fields)

        if start:
            self.start_plotting()            
            
    def __flatten_dict(dobj, kprefixs=[]):
        """
        Genenerator that can traverse through nested dictionaries and return
        key/value pairs

        For example given {'a':{'b':1}, 'c': 2}, it would yield
        ('a.b', 1) and ('c', 2)

        :param dobj dictionary object
        :param kprefixs  list of key of the directary and it's predecessors
        :yields key, value
        """
        sep = '.'
        for k, v in dobj.items():
            if type(v) == dict:
                yield from ScopeController.__flatten_dict(v, kprefixs + [k])
            else:
                yield sep.join(kprefixs + [k]), v

    def get_fdr(self):
        """
        Get EPICS7 PV field description back as a list

        :return: list of field description
        :raise PvaException: raise pvaccess exception when channel cannot be connected.
        """
        fdr = []
        fdr_scalar = []
        fdr_nonnumeric = []
        
        pv = self.model.get()
        
        if pv is None:
            return fdr, fdr_scalar, fdr_nonnumeric

        pv_structure = pv.getStructureDict()
        pv_dictionary = {k:v for k,v in ScopeController.__flatten_dict(pv_structure)}
        for k, v in pv_dictionary.items():
            if type(v) == list and all(type(e) == pva.ScalarType for e in v):
                if v[0] is pva.ScalarType.STRING:
                    fdr_nonnumeric.append(k)
                else:
                    fdr.append(k)
            elif type(v) == pva.ScalarType:
                fdr_scalar.append(k)
            if type(v) != np.ndarray:
                continue
            if len(v) == 0:
                continue

        fdr.sort()
        fdr_scalar.sort()
        return fdr, fdr_scalar, fdr_nonnumeric

    def update_fdr(self, empty=False):
        """
        Update EPICS7 PV field description

        :return:
        """
        if empty:
            fdr = []
            fdr_scalar = []
        else:
            try:
                fdr, fdr_scalar, fdr_nonnumeric = self.get_fdr()
            except pva.PvaException as e:
                self.notify_warning('Failed to get PV field description: ' + (str(e)))
                return

        fdr_all = fdr + fdr_nonnumeric
        fdr.insert(0, "None")
        fdr_scalar.insert(0, "None")

        # fill up the selectable pull down menu for array ID
        child = self.parameters.child("Config").child("ArrayId")
        child.setLimits(fdr_scalar)
        if child.value() != 'None':
            self.set_arrayid(child.value())
        
        # fill up the selectable pull down menu for x axes
        child = self.parameters.child("Config").child("X Axes")
        child.setLimits(fdr)
        if child.value() != 'None':
            self.set_xaxes(child.value())
                
        child = self.parameters.child("Trigger").child("Data Time Field")
        # Preserve current value before updating limits
        current_value = child.value()
        child.setLimits(fdr)
        # Restore value if it's still valid
        if current_value != 'None' and current_value in fdr:
            child.setValue(current_value)
        if child.value() != 'None':
            self._win.graphicsWidget.trigger.data_time_field = child.value()
        
        for idx in range(len(self.channels)):
            chan_name = "Channel %s" % (idx + 1)
            child = self.parameters.child(chan_name)
            c = child.child("Field")
            c.setLimits(fdr)
            if c.value() != 'None':
                self.set_channel_data(chan_name, 'Field', c.value())

        child = self.parameters.child('Config').child('Extra Display Fields')
        # Preserve current values before updating limits
        current_values = child.value()
        child.setLimits(fdr_all)
        # Restore values that are still valid
        if current_values:
            valid_values = [v for v in current_values if v in fdr_all]
            if valid_values:
                child.setValue(valid_values)
            
    def __failed_connection_callback(self, flag):
        """
        Called initially with flag=False if failed to connect to PV
        Will start periodically checking the connection.
        Once able to connect successfully, this function is called
        again with flag=True
        """
        if not flag:
            # Start periodically checking connection
            self.connection_timer.start(5000)
        else:
            # Got a connection, so turn off timer
            # and reload the fdr
            restart = self._win.graphicsWidget.plotting_started
            self.connection_timer.stop()
            self.stop_plotting()
            self.update_fdr()
            if restart:
                self.start_plotting()                
            
    def connection_changed(self, state, msg):
        for q in self.parameters.child("Acquisition").children():
            if q.name() == 'PV status':
                q.setValue(state)

        if state == 'Failed to connect':
            self.connection_timer_signal.sig.emit(False)

    def __check_connection(self):
        def success_callback(data):
            self.connection_timer_signal.sig.emit(True)

        self.model.async_get(success_callback=success_callback)

    def set_arrayid(self, value):
        """
        Set current field name for array id
            
        :param value:
        :return:
        """
        if value != self.current_arrayid:
            self.current_arrayid = value
            self._win.graphicsWidget.current_arrayid = value

    def set_xaxes(self, value):
        """
        Set current field name for x axes
            
        :param value:
        :return:
        """
        if value != self.current_xaxes:
            self.current_xaxes = value
            self._win.graphicsWidget.set_xaxes(value)

    def set_major_ticks(self, value):
        self._win.graphicsWidget.set_major_ticks(value)

    def set_minor_ticks(self, value):
        self._win.graphicsWidget.set_minor_ticks(value)


    def set_channel_data(self, channel_name, field, value):
        # avoid changes caused by Statistic updating
        for i, chan in enumerate(self.channels):
            if channel_name != 'Channel %i' % (i + 1):
                continue

            if field == 'Field':
                chan.pvname = value
            elif field == 'DC offset':
                chan.dc_offset = value
            elif field == 'Axis location':
                chan.axis_location = value
            
        self._win.graphicsWidget.setup_plot(channels=self.channels)
        
    def parameter_change(self, params, changes):
        """

        :param params:
        :param changes:
        :return:
        """
        for param, change, data in changes:
            if change == "value":
                path = self.parameters.childPath(param)
                if path is not None:
                    childName = '.'.join(path)
                else:
                    childName = param.name()

                if childName == "Acquisition.PV":
                    # stop DAQ and update pv info
                    try:
                        self.stop_plotting()
                        self.model.update_device(data, restart=False)
                        if data != "":
                            self.update_fdr()
                        else:
                            self.update_fdr(empty=True)
                            
                        # Recalculate object size and readjust buffer size
                        # if PV name changed
                        self.auto_buffer_size = True
                        self.object_size_tally = []
                        self.object_size = None
                    except Exception as e:
                        self.notify_warning('Failed to update PV: ' + (str(e)))
                    
                elif childName == "Acquisition.Start":
                    if data:
                        self.start_plotting()
                    else:
                        self.stop_plotting()
                elif "Channel" in childName:
                    chan, field = childName.split('.')
                    self.set_channel_data(chan, field, data)
                elif childName == "Config.ArrayId":
                    self.set_arrayid(data)
                elif childName == "Config.X Axes":
                    self.set_xaxes(data)
                elif childName == "Config.Major Ticks":
                    self.set_major_ticks(data)
                elif childName == "Config.Minor Ticks":
                    self.set_minor_ticks(data)
                elif childName == "Acquisition.Buffer Unit":
                    self.set_buffer_unit(data)
                elif 'Acquisition.Buffer' in childName:
                    self.auto_buffer_size = False
                    self.__calc_buffer_size()
                elif "Config.Extra Display Field" in childName:
                    self._win.graphicsWidget.set_mouseover_fields(data);
                elif "Config.MO Disp Location" in childName:
                    self._win.graphicsWidget.set_mouseover_display_location(data);
                elif childName == "Display.Mouse Over":
                    self._win.graphicsWidget.set_enable_mouseover(data)
        super().parameter_change(params, changes)
        

    def __calc_buffer_size(self):
        #
        # This function will adjust the number of samples to plot
        # based on buffer size, buffer unit, and object size
        # Called whenever one of these settings has changed
        #
        
        if self.buffer_unit == "Objects":
            if self.object_size:
                nobj = self.parameters.child("Acquisition").child("Buffer (Objects)").value()
                nsamples = nobj * self.object_size
                self._win.graphicsWidget.update_buffer(nsamples)
            
    def update_buffer_samples(self, size):
        """
        Sets number of samples in buffer

        :param size  number of samples
        """
        if self.buffer_unit == 'Samples':
            super().update_buffer_samples(size)
            return

        if self.buffer_unit == 'Objects':
            if self.object_size:
                nobj = math.ceil(size / self.object_size)
                self.parameters.child("Acquisition").child("Buffer (Objects)").setValue(nobj)
                self.__calc_buffer_size()
            else:
                self._win.graphicsWidget.update_buffer(size)
        else:
            raise Exception('Unknown buffer unit %s' % (self.buffer_unit))

    def set_buffer_unit(self, name):
        """
        Set units for buffer size.

        :param name buffer unit.  
        """
        if self.buffer_unit == name:
            return

        param = self.parameters.child("Acquisition").child("Buffer (%s)" % (self.buffer_unit))
        newname = "Buffer (%s)" % (name)
        param.setName(newname)        
        self.buffer_unit = name

        #Update buffer size based on current number of samples
        if self._win.graphicsWidget.max_length:
            self.update_buffer_samples(self._win.graphicsWidget.max_length)
        else:
            if self.buffer_unit == "Objects":
                nobj = self.parameters.child("Acquisition").child("Buffer (Objects)").value()
                if nobj == 0:
                    #default to 1 object
                    self.parameters.child("Acquisition").child("Buffer (Objects)").setValue(1)

                self.__calc_buffer_size()

    def set_object_size(self, size):
        """
        Set number of samples per object

        :param size number of samples per object
        """
        if size == self.object_size:
            return
        
        self.object_size = size
        self.__calc_buffer_size()
        
    def monitor_callback(self, data):
        # Calculate object size
        objlen = 0
        for k, v in ScopeController.__flatten_dict(dict(data)):
            try:
                objlen = max(len(v), objlen)
            except:
                pass

        if not self.object_size and objlen > 0:
            self.set_object_size(objlen)

            #Default buffer size to number of samples in an object
            #if buffer size was not explicitly set
            if self.auto_buffer_size:
                self.update_buffer_samples(self.object_size)

        self.object_size_tally.append(objlen)
        self.object_size_tally = self.object_size_tally[-5:]
        
        if not self._win.graphicsWidget.max_length:
            return
            
        def generator():
            yield from ScopeController.__flatten_dict(dict(data))
        self._win.graphicsWidget.data_process(generator)

    def start_plotting(self):
        """

        :return:
        """

        # stop a model first anyway to ensure it is clean
        self.model.stop()
        
        # start a new monitor
        self.model.start(self.monitor_callback)
        
        try:                
            super().start_plotting()
            self.parameters.child("Acquisition").child("Start").setValue(1)
        except Exception as e:
            self.parameters.child("Acquisition").child("Start").setValue(0)
            self.notify_warning('Failed to start plotting: ' + str(e))
            
    def stop_plotting(self):
        """

        :return:
        """
        super().stop_plotting()
        self.parameters.child("Acquisition").child("Start").setValue(0)
        # Stop data source
        self.model.stop()

    def update_status(self):
        super().update_status()

        if len(self.object_size_tally) > 0:
            avg_obj_size = statistics.mean(self.object_size_tally)
            self.parameters.child("Statistics").child('Avg Samples/Obj').setValue(avg_obj_size)
            self.set_object_size(math.ceil(avg_obj_size))

    def serialize(self, cfile):
        """
        Serialize the current scope configuration to an IO object.

        :param cfile: IO object (file-like) to write configuration to
        """
        from .config import Serializer, AppType, Scope

        serializer = Serializer()

        # Write app identifier
        serializer.set_app(AppType.SCOPE)

        # Serialize base scope configuration (Display, Acquisition, Trigger)
        # This includes buffer unit and buffer size
        self.serialize_scope_config(serializer)

        # Serialize scope-specific settings
        pv = self.parameters.child("Acquisition", "PV").value()
        if pv:
            serializer.set(Scope.PV, pv)

        start = self.parameters.child("Acquisition", "Start").value()
        serializer.set(Scope.CONNECT_ON_START, start)

        # Serialize Display settings
        mouse_over = self.parameters.child("Display", "Mouse Over").value()
        serializer.set(Scope.MOUSE_OVER, mouse_over)

        # Serialize Config settings
        arrayid = self.parameters.child("Config", "ArrayId").value()
        if arrayid and arrayid != "None":
            serializer.set(Scope.ARRAYID, arrayid)

        xaxes = self.parameters.child("Config", "X Axes").value()
        if xaxes and xaxes != "None":
            serializer.set(Scope.XAXES, xaxes)

        major_ticks = self.parameters.child("Config", "Major Ticks").value()
        if major_ticks:
            serializer.set(Scope.MAJORTICKS, major_ticks)

        minor_ticks = self.parameters.child("Config", "Minor Ticks").value()
        if minor_ticks:
            serializer.set(Scope.MINORTICKS, minor_ticks)

        extra_fields = self.parameters.child("Config", "Extra Display Fields").value()
        if extra_fields:
            serializer.set(Scope.EXTRA_DISPLAY_FIELDS, ','.join(extra_fields))

        mo_location = self.parameters.child("Config", "MO Disp Location").value()
        if mo_location:
            serializer.set(Scope.MOUSE_OVER_DISPLAY_LOCATION, mo_location)

        # Serialize channel configurations
        chan_cfgs = []
        for idx, channel in enumerate(self.channels):
            chan_num = idx + 1
            chan_param = self.parameters.child(f"Channel {chan_num}")

            field = chan_param.child("Field").value()
            if field and field != "None":
                chan_cfg = {
                    'field': field,
                    'color': channel.color
                }

                dc_offset = chan_param.child("DC offset").value()
                if dc_offset != 0:
                    chan_cfg['dcoffset'] = dc_offset

                axis_location = chan_param.child("Axis location").value()
                if axis_location:
                    chan_cfg['axislocation'] = axis_location

                chan_cfgs.append(chan_cfg)

        if chan_cfgs:
            serializer.write_channels('CHANNELS', chan_cfgs)

        # Write to file
        serializer.write(cfile)



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

from .pvapy_plugins import DataSource
from .pvapy_plugins import ConnectionState
from .pvapy_plugins import make_protocol



# -*- coding: utf-8 -*-

from .scope_config_base import ScopeConfigureBase
import re
from .config import Scope

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

PVA object viewer utilities

@author: Guobao Shen <gshen@anl.gov>
"""


class Configure(ScopeConfigureBase):
    """
    Scope application configuration panel settings
    """
    def __init__(self, params, **kwargs):
        """

        :param params: parameters parsed from command line and configuration file
        :param pvs: pv name dictionary, format: {"alias": "PV:Name"}
        """
        super().__init__(params, **kwargs)
        self.pvs = kwargs.get("pv", None)

        self.counts = 4
        self.default_color = ['#FFFF00', '#FF00FF', '#55FF55', '#00FFFF', '#5555FF',
                              '#5500FF', '#FF5555', '#0000FF', '#FFAA00', '#000000']
        
        self.default_arrayid = kwargs.get("arrayid", None)
        if not self.default_arrayid:
            self.default_arrayid = self.params.get(Scope.ARRAYID, 'None').capitalize()
        self.default_xaxes = kwargs.get("xaxes", "None")
        if not self.default_xaxes:
            self.default_xaxes = self.params.get(Scope.XAXES, 'None').capitalize()
        

        
    def assemble_channel(self, section=None):
        """
        Assemble channel information for plotting

        :param section:
        :return:
        """
        # get channel counts to display, 4 by default
        self.counts = self.params.get(Scope.CHANNEL_COUNT, default=4)
        channel = []

        #Read channel information.  Channel order is
        #determined by order in config
        chan_cfgs = self.params.get_channel_config()
        
        if len(chan_cfgs) > self.counts:
            self.counts = len(chan_cfgs)

        if self.counts > 10:
            # limit max channel to display
            self.counts = 10

        
        for i in range(self.counts):
            default_cfg = {
                'field' : 'None',
                'dcoffset': 0.0
            }
            
            chcfg = chan_cfgs[i] if len(chan_cfgs) > i else default_cfg
            field = chcfg.get('field', default_cfg['field'])
            dcoffset = float(chcfg.get('dcoffset', default_cfg['dcoffset']))

            channel.append(
                {"name": "Channel %s" % (i + 1),
                 "type": "group",
                 "children": [
                     {
                         "name": "Color",
                         "type": "color",
                         "value": self.default_color[i],
                         "readonly": True
                     },
                     {"name": "Field", "type": "list", "limits": [field], "value": field},
                     {"name": "DC offset", "type": "float", "value": dcoffset},
                     {"name": "Axis location", "type": "list", "limits": {
                         "Left" : "left",
                         "Right" : "right",
                     }, "value" : "Left"},
                 ]
                 }
            )

        return channel

    def assemble_display(self, section=None):
        display = super().assemble_display(section=section, app_section_key="SCOPE", default_autoscale=False)
        #FIXME: move this to base class once get mouseover support in striptool
        children = display['children']
        mouse_over = self.params.get(Scope.MOUSE_OVER, default=False)
        newchildren = [
            { "name" : "Mouse Over", "type": "bool", "value": mouse_over }
            ]
        children.extend(newchildren)
        
        return display

    def assemble_acquisition(self, section=None):
        buffer_unit = self.params.get(Scope.BUFFER_UNIT, default='samples')
        buffer_unit = buffer_unit.title()

        acquisition = super().assemble_acquisition(section, buffer_unit=buffer_unit)
        children = acquisition['children']

        start = self.params.get(Scope.CONNECT_ON_START, False)

        if self.pvs is not None:
            pv = list(self.pvs.values())[0]
        else:
            # it means PV map is not specified from command line
            # get one from configuration
            pv = self.params.get(Scope.PV)
            # if PV is available by default
            self.pvs = {pv: pv}

        newchildren = [
            {"name": "Buffer Unit", "type": "list", "limits": ["Samples", "Objects"],
             "value": buffer_unit},
            {"name": "PV", "type": "str", "value": pv},
            {"name": "PV status", "type": "str", "value": "Disconnected", "readonly": True},
            {"name": "Start", "type": "bool", "value": start}
        ]

        children.extend(newchildren)
        
        return acquisition

    def assemble_statistics(self):
        stats = super().assemble_statistics()
        children = stats['children']
        children.append( {"name": "Avg Samples/Obj", "type": "float", "value": 0, "readonly":True, "decimals":20})
        return stats
    
    def assemble_config(self, section=None):
        # Assemble extra configuration information for plotting
        # which is ArrayId selection, and x axes
        id_value = ["None"]
        if self.default_arrayid != "None":
            id_value.append(self.default_arrayid)

        axes = ["None"]
        if self.default_xaxes != "None":
            axes.append(self.default_xaxes)

        self.default_major_tick = self.params.get(Scope.MAJORTICKS, 0)
        self.default_minor_tick = self.params.get(Scope.MINORTICKS, 0)

        # Load extra display fields from config
        extra_fields = self.params.get(Scope.EXTRA_DISPLAY_FIELDS, [])

        mo_display_loc = self.params.get(Scope.MOUSE_OVER_DISPLAY_LOCATION, 'bottom_right').replace('_', '-')

        cfg = {"name": "Config",
               "type": "group",
               "expanded": True,
               "children": [
                   {"name": "ArrayId", "type": "list", "limits": id_value, "value": self.default_arrayid},
                   {"name": "X Axes", "type": "list", "limits": axes, "value": self.default_xaxes},
                   {"name": "Major Ticks", "type": "int", "value": self.default_major_tick, 'decimals':20},
                   {"name": "Minor Ticks", "type": "int", "value": self.default_minor_tick, 'decimals':20},
                   {"name": "Extra Display Fields", "type": "checklist", "value": extra_fields, "limits": extra_fields, "expanded": False},
                   {"name": "MO Disp Location", "type": "list", "limits": ['top-right', 'bottom-right', 'bottom-left'], "value": mo_display_loc}

                   ]
               }
        return cfg

    def parse(self):
        """

        :return:
        """
        
        acquisition = self.assemble_acquisition()
        display = self.assemble_display()
        channel = self.assemble_channel()
        trigger = self.assemble_trigger()
        cfg = self.assemble_config()

        # line up in order
        paramcfg = [acquisition, trigger, display, cfg]
        for ch in channel:
            paramcfg.append(ch)
        statistics = self.assemble_statistics()
        
        paramcfg.append(statistics)

        return paramcfg



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

import unittest
import logging
from configparser import ConfigParser
from pprint import pformat
from io import StringIO

from c2dataviewer.control.scopeconfig import Configure

# ----------------------------------------------------------------------
# Configure logging to capture debug messages in a buffer. The messages
# will only be printed if a test fails, making the output cleaner for
# successful test runs.
# ----------------------------------------------------------------------
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class TestScopeConfig(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Set up logging capture once for all tests."""
        cls.log_stream = StringIO()
        cls.log_handler = logging.StreamHandler(cls.log_stream)
        cls.log_handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(levelname)s:%(name)s:%(message)s')
        cls.log_handler.setFormatter(formatter)
        logger.addHandler(cls.log_handler)

    @classmethod
    def tearDownClass(cls):
        """Clean up logging after all tests."""
        logger.removeHandler(cls.log_handler)
        cls.log_stream.close()

    def get_display_val(self, data, val):
        """
        Helper method to extract the value for a given entry (by name) from
        the children list returned by Configure helper methods.

        Emits debug information whenever a value is looked-up so that the
        test output clearly indicates what is being searched and (when
        found) what value is returned.
        """
        logger.debug("Looking for key '%s' in data:\n%s", val, pformat(data))
        for child in data['children']:
            if child['name'] == val:
                logger.debug("Key '%s' found with value: %s", val, child['value'])
                return child['value']
        logger.debug("Key '%s' not found!", val)
        return None

    def dump_config(self, config_parser):
        """
        Dumps content from a ConfigParser object.
        
        Args:
        config_parser: A ConfigParser instance
        
        Returns:
        str: String representation of the config
        """
        result = []
        for section in config_parser.sections():
            result.append(f"[{section}]")
            for key, value in config_parser.items(section):
                result.append(f"{key} = {value}")
            result.append("")
            
        return "\n".join(result)

    # ------------------------------------------------------------------
    # Existing tests
    # ------------------------------------------------------------------
    def test_trigger(self):
        raw1 = """
        [TRIGGER]
        TRIGGER=pv1
        TRIGGER_MODE=onchange
        """
        parser = ConfigParser()
        parser.read_string(raw1)
        configure = Configure(parser)
        section = parser["TRIGGER"]
        self.assertIsNotNone(section)
        trigger = configure.assemble_trigger(section=section)

        logger.debug("Assembled trigger structure:\n%s", pformat(trigger))

        self.assertEqual(self.get_display_val(data=trigger, val='PV'), 'pv1')
        self.assertEqual(self.get_display_val(data=trigger, val='Mode'), 'onchange')

    def test_display_configuration(self):
        """
        Comprehensive test for display configuration including:
        - Autoscale precedence (SCOPE section overrides DISPLAY section)
        - All settings with valid values
        - Default values when settings absent
        - Bad/invalid input handling with fallback to defaults
        """
        # Test 1: Autoscale precedence - SCOPE section overrides DISPLAY
        raw1 = """
        [SCOPE]
        DefaultProtocol = ca
        AUTOSCALE=False

        [DISPLAY]
        AUTOSCALE=True
        AVERAGE=1
        HISTOGRAM=False
        N_BIN=100
        REFRESH=100
        """
        parser = ConfigParser()
        parser.read_string(raw1)
        configure = Configure(parser)
        logger.debug('CONFIG:\n%s', self.dump_config(parser))
        display = configure.assemble_display()
        logger.debug("Assembled display (autoscale precedence):\n%s", pformat(display))
        # SCOPE section has AUTOSCALE=False, so it takes precedence
        self.assertFalse(self.get_display_val(data=display, val='Autoscale'))

        # Test 2: Autoscale from DISPLAY section when not in SCOPE
        raw2 = """
        [SCOPE]
        DefaultProtocol = ca

        [DISPLAY]
        AUTOSCALE=True
        """
        parser = ConfigParser()
        parser.read_string(raw2)
        configure = Configure(parser)
        section = parser["DISPLAY"]
        display = configure.assemble_display(section=section)
        logger.debug("Assembled display (autoscale from display):\n%s", pformat(display))
        self.assertTrue(self.get_display_val(data=display, val='Autoscale'))

        # Test 3: All settings with valid values
        raw3 = """
        [SCOPE]
        DefaultProtocol = ca
        AUTOSCALE=True

        [DISPLAY]
        MODE=fft
        FFT_FILTER=hamming
        AVERAGE=10
        SINGLE_AXIS=False
        HISTOGRAM=True
        N_BIN=256
        REFRESH=500
        """
        parser = ConfigParser()
        parser.read_string(raw3)
        configure = Configure(parser)
        section = parser["DISPLAY"]
        display = configure.assemble_display(section=section)
        logger.debug("Assembled display (all settings):\n%s", pformat(display))
        self.assertEqual(self.get_display_val(display, 'Mode'), 'fft')
        self.assertEqual(self.get_display_val(display, 'FFT filter'), 'hamming')
        self.assertEqual(self.get_display_val(display, 'Exp moving avg'), 10)
        self.assertTrue(self.get_display_val(display, 'Autoscale'))  # SCOPE precedence
        self.assertFalse(self.get_display_val(display, 'Single axis'))
        self.assertTrue(self.get_display_val(display, 'Histogram'))
        self.assertEqual(self.get_display_val(display, 'Num Bins'), 256)
        self.assertEqual(self.get_display_val(display, 'Refresh'), 0.5)  # 500ms -> 0.5s

        # Test 4: Default values when no DISPLAY section
        raw4 = """
        [SCOPE]
        DefaultProtocol = ca
        """
        parser = ConfigParser()
        parser.read_string(raw4)
        configure = Configure(parser)
        display = configure.assemble_display(section=None)
        logger.debug("Assembled display (defaults):\n%s", pformat(display))
        self.assertEqual(self.get_display_val(display, 'Mode'), 'normal')
        self.assertEqual(self.get_display_val(display, 'FFT filter'), 'none')
        self.assertEqual(self.get_display_val(display, 'Exp moving avg'), 1)
        self.assertFalse(self.get_display_val(display, 'Autoscale'))
        self.assertTrue(self.get_display_val(display, 'Single axis'))
        self.assertFalse(self.get_display_val(display, 'Histogram'))
        self.assertEqual(self.get_display_val(display, 'Num Bins'), 100)
        self.assertEqual(self.get_display_val(display, 'Refresh'), 0.1)

        # Test 5: Bad/invalid input falls back to defaults
        raw5 = """
        [DISPLAY]
        MODE=invalidmode
        FFT_FILTER=bogusfilter
        AVERAGE=-5
        SINGLE_AXIS=maybe
        HISTOGRAM=maybe
        N_BIN=badint
        REFRESH=badfloat
        """
        parser = ConfigParser()
        parser.read_string(raw5)
        configure = Configure(parser)
        section = parser["DISPLAY"]
        display = configure.assemble_display(section=section)
        logger.debug("Assembled display (bad input):\n%s", pformat(display))
        self.assertEqual(self.get_display_val(display, 'Mode'), 'normal')
        self.assertEqual(self.get_display_val(display, 'FFT filter'), 'none')
        self.assertEqual(self.get_display_val(display, 'Exp moving avg'), 1)
        self.assertTrue(self.get_display_val(display, 'Single axis'))
        self.assertFalse(self.get_display_val(display, 'Histogram'))
        self.assertEqual(self.get_display_val(display, 'Num Bins'), 100)
        self.assertEqual(self.get_display_val(display, 'Refresh'), 0.1)

    # ------------------------------------------------------------------
    # Channel configuration tests
    # ------------------------------------------------------------------
    def test_channel_configuration(self):
        """
        Comprehensive test for channel configuration including:
        - Basic channel definitions with fields and offsets
        - Default values when no channels specified
        - Channel count override when more channels defined than COUNT
        - Maximum channel limit (10 channels)
        - Partial configuration with mixed defined/default channels
        """
        # Test 1: Basic channel configuration
        raw1 = """
        [SCOPE]
        DefaultProtocol = ca

        [CHANNELS]
        COUNT=2
        chan0.field=field1
        chan0.dcoffset=1.5
        chan1.field=field2
        chan1.dcoffset=-2.0
        """
        parser = ConfigParser()
        parser.read_string(raw1)
        configure = Configure(parser)
        channels = configure.assemble_channel()
        logger.debug("Assembled channels (basic):\n%s", pformat(channels))
        self.assertEqual(len(channels), 2)
        self.assertEqual(channels[0]['name'], 'Channel 1')
        self.assertEqual(self.get_display_val(channels[0], 'Field'), 'field1')
        self.assertEqual(self.get_display_val(channels[0], 'DC offset'), 1.5)
        self.assertEqual(self.get_display_val(channels[0], 'Color'), '#FFFF00')
        self.assertEqual(channels[1]['name'], 'Channel 2')
        self.assertEqual(self.get_display_val(channels[1], 'Field'), 'field2')
        self.assertEqual(self.get_display_val(channels[1], 'DC offset'), -2.0)
        self.assertEqual(self.get_display_val(channels[1], 'Color'), '#FF00FF')

        # Test 2: Default values when no channels specified
        raw2 = """
        [SCOPE]
        DefaultProtocol = ca
        """
        parser = ConfigParser()
        parser.read_string(raw2)
        configure = Configure(parser)
        channels = configure.assemble_channel()
        logger.debug("Assembled channels (defaults):\n%s", pformat(channels))
        self.assertEqual(len(channels), 4)  # 4 channels by default
        for i, ch in enumerate(channels):
            self.assertEqual(ch['name'], f'Channel {i+1}')
            self.assertEqual(self.get_display_val(ch, 'Field'), 'None')
            self.assertEqual(self.get_display_val(ch, 'DC offset'), 0.0)

        # Test 3: Channel count override when more channels defined
        raw3 = """
        [SCOPE]
        DefaultProtocol = ca

        [CHANNELS]
        COUNT=2
        chan0.field=field1
        chan1.field=field2
        chan2.field=field3
        chan3.field=field4
        chan4.field=field5
        """
        parser = ConfigParser()
        parser.read_string(raw3)
        configure = Configure(parser)
        channels = configure.assemble_channel()
        logger.debug("Assembled channels (count override):\n%s", pformat(channels))
        self.assertEqual(len(channels), 5)  # Overrides COUNT=2
        self.assertEqual(self.get_display_val(channels[0], 'Field'), 'field1')
        self.assertEqual(self.get_display_val(channels[4], 'Field'), 'field5')

        # Test 4: Maximum channel limit
        raw4 = """
        [SCOPE]
        DefaultProtocol = ca

        [CHANNELS]
        COUNT=15
        """
        parser = ConfigParser()
        parser.read_string(raw4)
        configure = Configure(parser)
        channels = configure.assemble_channel()
        logger.debug("Assembled channels (max limit):\n%s", pformat(channels))
        self.assertEqual(len(channels), 10)  # Limited to 10 channels

        # Test 5: Partial channel configuration
        raw5 = """
        [SCOPE]
        DefaultProtocol = ca

        [CHANNELS]
        COUNT=4
        chan0.field=field1
        chan0.dcoffset=5.0
        chan2.field=field3
        """
        parser = ConfigParser()
        parser.read_string(raw5)
        configure = Configure(parser)
        channels = configure.assemble_channel()
        logger.debug("Assembled channels (partial config):\n%s", pformat(channels))
        self.assertEqual(len(channels), 4)
        # Channels from config appear first in order defined
        self.assertEqual(self.get_display_val(channels[0], 'Field'), 'field1')
        self.assertEqual(self.get_display_val(channels[0], 'DC offset'), 5.0)
        self.assertEqual(self.get_display_val(channels[1], 'Field'), 'field3')
        self.assertEqual(self.get_display_val(channels[1], 'DC offset'), 0.0)
        # Remaining channels use defaults
        self.assertEqual(self.get_display_val(channels[2], 'Field'), 'None')
        self.assertEqual(self.get_display_val(channels[3], 'Field'), 'None')

    def test_channel_properties(self):
        """
        Test specific channel properties including:
        - Color assignment from palette
        - DC offset with various numeric formats
        - Axis location defaults
        """
        # Test 1: Color assignment
        raw1 = """
        [SCOPE]
        DefaultProtocol = ca

        [CHANNELS]
        COUNT=5
        """
        parser = ConfigParser()
        parser.read_string(raw1)
        configure = Configure(parser)
        channels = configure.assemble_channel()
        logger.debug("Assembled channels (colors):\n%s", pformat(channels))
        expected_colors = ['#FFFF00', '#FF00FF', '#55FF55', '#00FFFF', '#5555FF']
        for i, (ch, expected_color) in enumerate(zip(channels, expected_colors)):
            color = self.get_display_val(ch, 'Color')
            self.assertEqual(color, expected_color,
                           f"Channel {i} color mismatch: expected {expected_color}, got {color}")

        # Test 2: DC offset with different numeric formats
        raw2 = """
        [SCOPE]
        DefaultProtocol = ca

        [CHANNELS]
        COUNT=4
        chan0.dcoffset=0
        chan1.dcoffset=10.5
        chan2.dcoffset=-5.25
        chan3.dcoffset=1e-3
        """
        parser = ConfigParser()
        parser.read_string(raw2)
        configure = Configure(parser)
        channels = configure.assemble_channel()
        logger.debug("Assembled channels (dcoffset types):\n%s", pformat(channels))
        self.assertEqual(self.get_display_val(channels[0], 'DC offset'), 0.0)
        self.assertEqual(self.get_display_val(channels[1], 'DC offset'), 10.5)
        self.assertEqual(self.get_display_val(channels[2], 'DC offset'), -5.25)
        self.assertAlmostEqual(self.get_display_val(channels[3], 'DC offset'), 0.001)

        # Test 3: Axis location defaults
        raw3 = """
        [SCOPE]
        DefaultProtocol = ca

        [CHANNELS]
        COUNT=2
        chan0.field=field1
        chan1.field=field2
        """
        parser = ConfigParser()
        parser.read_string(raw3)
        configure = Configure(parser)
        channels = configure.assemble_channel()
        logger.debug("Assembled channels (axis location):\n%s", pformat(channels))
        for i, ch in enumerate(channels):
            axis_loc = self.get_display_val(ch, 'Axis location')
            self.assertEqual(axis_loc, 'Left',
                           f"Channel {i} should have default axis location 'Left'")


if __name__ == '__main__':
    unittest.main()



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""
from .pvconfig import PvConfig
import re
from .scope_config_base import ScopeConfigureBase
from .config import Striptool
import logging

def parse_pv(val):
    pvname = None
    if "://" in val:
        proto, pvname = val.split('://')
        return pvname, proto
    else:
        pvname = val
        return pvname, None

class StripToolConfigure(ScopeConfigureBase):
    def __init__(self, params, **kwargs):
        super().__init__(params, **kwargs)
        self.default_proto = None
        self.pvs = {}

        try:
            cfg = self.params.cfg['STRIPTOOL']
        except:
            return

        self.default_proto = self.params.get(Striptool.DEFAULT_PROTOCOL, default='ca')

        cfgpvs = self.params.get_channel_config(is_striptool=True)
        for cfg in cfgpvs:
            chcfg = PvConfig()
            for k, v in cfg.items():
                if k == 'pv':
                    pvname, proto = parse_pv(v)
                    chcfg.pvname = pvname
                    if proto is None:
                        proto = self.default_proto
                    chcfg.set_proto(proto)
                elif k == 'color':
                    chcfg.color = str(v)
            
            if chcfg.pvname:
                self.pvs[chcfg.pvname] = chcfg
            
        pvs = kwargs.get('pv')
        if pvs:
            for pv in pvs.values():
                chcfg = None
                pvname, proto = parse_pv(pv)
                if pvname in self.pvs:
                    chcfg = self.pvs[pvname]
                else:
                    chcfg = PvConfig()
                    chcfg.set_proto(self.default_proto)
                    
                chcfg.pvname = pvname
                if proto:
                    chcfg.set_proto(proto)
                self.pvs[pvname] = chcfg                

    def assemble_acquisition(self, section=None):
        acquisition = super().assemble_acquisition(section)
        children = acquisition['children']

        # Load Sample Mode from config
        sample_mode = self.params.get(Striptool.SAMPLEMODE, default=True)
        # Handle string values from config file
        if isinstance(sample_mode, str):
            sample_mode = sample_mode.lower() in ('true', '1', 'yes')

        children.append({"name": "Sample Mode", "type":"bool", "value": sample_mode})

        return acquisition
    
    def assemble_display(self, section=None):
        display = super().assemble_display(section=section, app_section_key="STRIPTOOL", default_autoscale=True)

        return display
    
    def parse(self):
        acquisition = self.assemble_acquisition()
        display = self.assemble_display()            
        statistics = self.assemble_statistics()
        # line up in order
        paramcfg = [acquisition, display, statistics]

        return paramcfg


